#!/usr/bin/perl -w

# Program fatparser.pl

# (c) Gary C. Kessler, 2011, 2014 [gck@garykessler.net]

# Program to parse FAT directory data structures.
# Ref: File System Forensic Analysis, Brian Carrier

# ********************************************************
# MAIN PROGRAM BLOCK
# ********************************************************

use File::Copy qw(move);

# Initialize global variables

$version = "2.1";
$build_date = "01 July 2016";

# Must obtain source file name and FAT type.

$source = "";
$fat_type = "";
$dump_mode = 0;
$csv = 0;

#
# Begin program activity!!
#

print "\nFAT Table Parser v$version - Gary C. Kessler ($build_date)\n\n";

# Parse command for file name(s) and request missing information.

if (parsecommandline ())
  { exit 1; };

# If requested, ensure that we can open/create the output file

if ($csv == 1)
  {
  if (ofh ("OUTFILE",$csv_file))
    { exit 1; };
  print "CSV output file: $csv_file\n";
  };

# Open input ($source) file and stop program if error

open (INFILE, "<:raw", $source)
or die "Can't open input file \"$source\": $!";
binmode INFILE;

if ($dump_mode == 1)
  {
  print "Source file = $source (FAT type: $fat_type) FILE DUMP\n\n";
  dump_mode ();
  exit 1;
  }

# Make a guess as to the FAT type, then "rewind" the input file.

if ($fat_type eq "FAT0")
  {
  determine_FAT_type ();

  if ($fat_type eq "XXX")
    {
    print "Cannot determine FAT type...\n\n";
    exit 1;
    }

  seek (INFILE,0,0);
  }

# Determine media descriptor and EOC.

determine_FAT_info ();

printf "Source file = %s (FAT type: %s, ", $source, $fat_type;

# print header information and then process rest of the FAT structure

if ($fat_type eq "FAT12")
    {
    printf "FAT[0]: 0x%03X, FAT[1]: 0x%03X)\n\n", $media_type, $eoc;
    if ($csv == 0)
        { parse_FAT12 (); }
      else
        { parse_FAT12_csv (); }
    }
  else
    {
    if ($fat_type eq "FAT16")
        {
        printf "FAT[0]: 0x%04X, FAT[1]: 0x%04X)\n\n", $media_type, $eoc;
        $entry_len = 2;
        if ($csv == 0)
            { parse_BIGFAT (); }
          else
            { parse_BIGFAT_csv (); }
        }
      else
        {
        printf "FAT[0]: 0x%08X, FAT[1]: 0x%08X)\n\n", $media_type, $eoc;
        $entry_len = 4;
        if ($csv == 0)
            { parse_BIGFAT (); }
          else
            { parse_BIGFAT_csv (); }
        }
    }

close INFILE;
if ($csv == 1) { close OUTFILE; };
print "\nDone!\n";


# ********************************************************
# *****           SUBROUTINE BLOCKS                  *****
# ********************************************************

# ********************************************************
# determine_FAT_info
# ********************************************************

# Read the starting entries of the FAT structure to determine FAT[0],
# which contains the media descriptor, and FAT[1] which is the EOC mark.
#
# Have to handle FAT12, FAT16, and FAT32 slightly differently....

sub determine_FAT_info
{
local ($i, $tmp);

FAT_TYPE:
  {
  $fat_type eq "FAT12" && do
    {
    for ($i = 0; $i <= 2; $i++)
      {
      read (INFILE,$byte,1);
      $tmp [$i] = unpack ("C", $byte);
      }
    $media_type = int (($tmp [1] % 16) * 256 + $tmp [0]);
    $eoc = int ($tmp [2] * 16 + ($tmp [1] / 16));
    };

  $fat_type eq "FAT16" && do
    {
    for ($i = 0; $i <= 1; $i++)
      {
      read (INFILE,$byte,1);
      $fat [$i] = unpack ("C", $byte);
      }
    $media_type = little_endian (0,2,@fat);
    for ($i = 0; $i <= 1; $i++)
      {
      read (INFILE,$byte,1);
      $fat [$i] = unpack ("C", $byte);
      }
    $eoc = little_endian (0,2,@fat);
    };

  $fat_type eq "FAT32" && do
    {
    for ($i = 0; $i <= 3; $i++)
      {
      read (INFILE,$byte,1);
      $fat [$i] = unpack ("C", $byte);
      }
    $media_type = little_endian (0,4,@fat);
    for ($i = 0; $i <= 3; $i++)
      {
      read (INFILE,$byte,1);
      $fat [$i] = unpack ("C", $byte);
      }
    $eoc = little_endian (0,4,@fat);
    };
  }

return;
}

# ********************************************************
# determine_FAT_type
# ********************************************************

# Read the first few bytes to determine the FAT type. In general:
#   - FAT12 starts with F0 FF FF
#   - FAT16 starts with F8 FF FF FF
#   - FAT32 starts with F8 FF FF 0F

sub determine_FAT_type
{
local ($i, $tmp);

# Read and examine first byte

read (INFILE,$byte,1);
$tmp = unpack ("C", $byte);

if ($tmp == 0xF0)
    {
    $fat_type = "FAT12";
    return;
    }
  else
    {
    if ($tmp != 0xF8)
      {
      $fat_type = "XXX";
      return;
      }
    }

# Now get to 4th byte

for ($i = 0; $i <= 2; $i++)
  { read (INFILE,$byte,1); }

$tmp = unpack ("C", $byte);

if ($tmp == 15)
    { $fat_type = "FAT32"; }
  else
    {
    if ($tmp == 0xFF)
        { $fat_type = "FAT16"; }
      else
        { $fat_type = "XXX" ; }
    }

return;
}

# ********************************************************
# dump_mode
# ********************************************************

# Print out the raw FAT file

sub dump_mode
{
local ($i, $tmp);

print "**OFFSET**";
$i = 0;

while (not (eof INFILE))
  {
  if ($i % 16 == 0)
    { printf "\n[%08X] ", $i; }
  read (INFILE,$byte,1);
  $tmp = unpack ("C", $byte);
  printf " %02X", $tmp;
  $i++;
  }

print "\n\n";
return;
}

# ********************************************************
# help_text
# ********************************************************

# Display the help file

sub help_text
{
print<< "EOT";
Program usage: fatparser [-i input_file] [-12 | -16 | -32 | -0] {-d} {-c output_file}
               fatparser {[-h] | [-t]}

 where: -i is the input file name
        -12, -16, or -32 indicates the FAT type
        -c requests CSV output to a file
        -0 asks the program to guess the FAT type
        -d asks for a raw dump of the FAT table
        -h prints this help file
        -t prints the FAT template file

If no -i switch is provided, the program will ask for the input
file name.

If no FAT type switch is provided, the program will ask for the
FAT type.

The input file should contain the contents of a FAT structure.
The best way to obtain this is to export the FAT table from your
forensics software.

Consider piping the output to 'more' when using FAT16 or FAT32.

EOT
return;
}

# ********************************************************
# little_endian
# ********************************************************

# Given an array name, starting position, and number of digits,
# calculate the decimal value using little endian (LSB first)

sub little_endian
{
my ($start,$n,@array) = @_;
my ($i, $sum);

$sum = $array[$start];
for ($i=1; $i<$n; $i++)
  { $sum += $array[$start+$i] * (2**($i*8)); }

return $sum;
}

# ********************************************************
# ofh
# ********************************************************

# Output File Handler

# When requested to output to a file, check to see if file
# exists. If not, open it. If it does, ask the user what
# action to take.

# Usage: ofh (filehandle, filename)
# NOTE:  Requires that line "use File::Copy qw(move);" appear in main

# Return $state = 1 to indicate that the program should exit

sub ofh
{
my ($filehandle, $filename) = @_;
local ($action, $state);

if (-e $filename)
    {
    print "$filename exists! What to do?";
    do
      {
      print "\na = append, b = backup old file, o = overwrite, x = exit: ";
      chomp ($action = <STDIN>);
      $action = lc ($action);
      }
      until ($action eq "a" || $action eq "b" || $action eq "o" || $action eq "x");

    print "\n";
    }
  else
    { $action = "o"; };

# Take action as requested above

$state = 0;
ACT:
  {
  $action eq "a" && do
    {
    open ($filehandle, ">>", $filename)
    or die "Can't open output file \"$filename\": $!";
    return $state;
    };
  $action eq "b" && do
    {
    move $filename, $filename . ".bak"
    or die "Can't create backup file \"$filename\".bak: $!";
    open ($filehandle, ">", $filename)
    or die "Can't open output file \"$filename\": $!";
    return $state;
    };
  $action eq "o" && do
    {
    open ($filehandle, ">", $filename)
    or die "Can't open output file \"$filename\": $!";
    return $state;
    };
  $action eq "x" && do
    {
    $state = 1;
    return $state;
    };
  };
};

# ********************************************************
# parsecommandline
# ********************************************************

# Parse command line for file name, if present. Query
# user for any missing information

# Return $state = 1 to indicate that the program should stop
# immediately (switch -h)

sub parsecommandline
{
my $state = 0;
local $temp;

# Parse command line switches ($ARGV array of length $#ARGV)

if ($#ARGV >= 0)
  { 
  for ($i = 0; $i <= $#ARGV; $i++)
    {
    PARMS:
      {
      $ARGV[$i] eq "-i" && do
         {
         $source = $ARGV[$i+1];
         $i++;
         last PARMS;
         };
      $ARGV[$i] eq "-c" && do
         {
         $csv = 1;
         $csv_file = $ARGV[$i+1];
         $i++;
         last PARMS;
         };
      $ARGV[$i] eq "-t" && do
         {
         $layout = "layout_fat.txt";
         open (INFILE, "<", $layout)
          or die "Can't open input file \"$layout\": $!";
         while (not (eof INFILE))
           {
           $line = <INFILE>;
           print $line;
           }
         close INFILE;
         $state = 1;
         return $state;
         };
      $ARGV[$i] eq "-d" && do
         {
         $dump_mode = 1;
         last PARMS;
         };
      $ARGV[$i] eq "-12" && do
         {
         $fat_type = "FAT12";
         last PARMS;
         };
      $ARGV[$i] eq "-16" && do
         {
         $fat_type = "FAT16";
         last PARMS;
         };
      $ARGV[$i] eq "-32" && do
         {
         $fat_type = "FAT32";
         last PARMS;
         };
      $ARGV[$i] eq "-0" && do
         {
         $fat_type = "FAT0";
         last PARMS;
         };
      $ARGV[$i] eq "-h" && do
         {
         help_text();
         $state = 1;
         return $state;
         };

      do
         {
         print "Invalid parameter \"$ARGV[$i]\".\n\n";
         print "Usage: fatparser [-i input_file] [-12 | -16 | -32 | -0] {-d}\n";
         print "       fatparser {[-h] | [-t]}\n\n";
         $state = 1;
         return $state;
         };
      };
    };
  };

# Prompt for file name and FAT type, if not supplied

if ($source eq "")
  {
  print "Enter the input file name: ";
  chomp ($source = <STDIN>);
  print "\n";
  };

if ($fat_type eq "")
  {
  print "Enter the FAT type (12, 16, 32 or 0): ";
  chomp ($temp = <STDIN>);
  print "\n";
  if ($temp != 12 && $temp != 16 && $temp != 32 && $temp != 0)
      { $state = 1; }
    else
      { $fat_type = "FAT" . $temp; }
  };

return $state;
}

# ********************************************************
# parse_BIGFAT
# ********************************************************

# Parse a FAT16 or FAT32 table and display the entries.

sub parse_BIGFAT
{
local ($e, $entry_len, $fat, $fat_entry, $i, $j, $value, $width);

# Set variables based upon FAT type, and print FAT[0] and FAT[1]

if ($fat_type eq "FAT16")
    {
    $entry_len = 2;
    printf ("[         0] =>      x%04X  [         1] =>      x%04X  ", $media_type, $eoc);
    }
  else
    {
    $entry_len = 4;
    printf ("[         0] =>  x%08X  [         1] =>  x%08X  ", $media_type, $eoc);
    }

$width = 3;
$fat_entry = 2;


while (not (eof INFILE))
  {
  for ($i = 0; $i < $entry_len; $i++)
    {
    if (eof INFILE)
        {
        for ($j = $i; $j < $entry_len; $j++)
          { $fat [$j] = 0; }
        last;
        }
      else
        {
        read (INFILE,$byte,1);
        $fat [$i] = unpack ("C", $byte);
        }
    }

  $value = little_endian (0,$entry_len,@fat);

  printf ("[%10d] => ", $fat_entry);
    if ($value == $eoc )
        { print "EOC       "; }
      else
        { printf ("%10d", $value); }
  
    $fat_entry++;
    if ($fat_entry % $width == 0)
        { print "\n"; }
      else
        { print "  "; }
  }

print "\n\n";
return;
}

# ********************************************************
# parse_BIGFAT_csv
# ********************************************************

# Parse a FAT16 or FAT32 table and write entries to a file.
# Two subroutines to avoid a huge mess o' code!!!

sub parse_BIGFAT_csv
{
local ($e, $entry_len, $fat, $fat_entry, $i, $j, $value, $width);

# Set variables based upon FAT type, and print FAT[0] and FAT[1]

if ($fat_type eq "FAT16")
    { $entry_len = 2; }
  else
    { $entry_len = 4; }

print OUTFILE "0, $media_type\n";
print OUTFILE "1, $eoc\n";

$fat_entry = 2;

while (not (eof INFILE))
  {
  for ($i = 0; $i < $entry_len; $i++)
    {
    if (eof INFILE)
        {
        for ($j = $i; $j < $entry_len; $j++)
          { $fat [$j] = 0; }
        last;
        }
      else
        {
        read (INFILE,$byte,1);
        $fat [$i] = unpack ("C", $byte);
        }
    }

  $value = little_endian (0,$entry_len,@fat);

  if ($value == $eoc )
      { print OUTFILE "$fat_entry,EOC\n"; }
    else
      { print OUTFILE "$fat_entry,$value\n"; }
  $fat_entry++;
  }
return;
}

# ********************************************************
# parse_FAT12
# ********************************************************

# Parse the FAT12 table and display the entries.
# In FAT12, two entries are packed into three bytes.

sub parse_FAT12
{
local ($e, $entry_len, $fat, $fat_entry, $i, $j, $value);

# Start by printing FAT[0] and FAT[1]

printf ("[0000] => x%03X  [0001] => x%03X  ", $media_type, $eoc);
$fat_entry = 2;
$entry_len = 3;
$width = 5;

while (not (eof INFILE))
  {
  for ($i = 0; $i < $entry_len; $i++)
    {
    if (eof INFILE)
        {
        for ($j = $i; $j < $entry_len; $j++)
          { $fat [$j] = 0; }
        last;
        }
      else
        {
        read (INFILE,$byte,1);
        $fat [$i] = unpack ("C", $byte);
        }
    }

  $e [0] = int (($fat [1] % 16) * 256 + $fat [0]);
  $e [1] = int ($fat [2] * 16 + ($fat [1] / 16));

  for ($i = 0; $i < 2; $i++)
    {
    $value = $e [$i];
    printf ("[%04d] => ", $fat_entry);
    if ($value == $eoc)
        { print "EOC "; }
      else
        { printf ("%04d", $value); }

    $fat_entry++;
    if ($fat_entry % $width == 0)
        { print "\n"; }
      else
        { print "  "; }
    }
  }

print "\n\n";
return;
}

# ********************************************************
# parse_FAT12_csv
# ********************************************************

# Parse the FAT12 table and write entries to a CSV file.
# In FAT12, two entries are packed into three bytes.

sub parse_FAT12_csv
{
local ($e, $entry_len, $fat, $fat_entry, $i, $j, $value);

# Start by printing FAT[0] and FAT[1]

print OUTFILE "0, $media_type\n";
print OUTFILE "1, $eoc\n";

$fat_entry = 2;
$entry_len = 3;

while (not (eof INFILE))
  {
  for ($i = 0; $i < $entry_len; $i++)
    {
    if (eof INFILE)
        {
        for ($j = $i; $j < $entry_len; $j++)
          { $fat [$j] = 0; }
        last;
        }
      else
        {
        read (INFILE,$byte,1);
        $fat [$i] = unpack ("C", $byte);
        }
    }

  $e [0] = int (($fat [1] % 16) * 256 + $fat [0]);
  $e [1] = int ($fat [2] * 16 + ($fat [1] / 16));

  for ($i = 0; $i < 2; $i++)
    {
    $value = $e [$i];
    if ($value == $eoc )
      { print OUTFILE "$fat_entry,EOC\n"; }
    else
      { print OUTFILE "$fat_entry,$value\n"; }

    $fat_entry++;
    }
  }

return;
}

